//
//	16bit Color LCD用 グラフィックファイル作成ツール
//	24bit BMP ファイルを GRPファイル(独自フォーマット)に変換する
//
//	2011.5.31	KAWAKAMI Yukio
//

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<ctype.h>

#define	WIDTH_MAX	1024
#define	HEIGHT_MAX	1024

#pragma pack(1)
typedef struct tagBITMAPFILEHEADER {
  unsigned short bfType;
  unsigned long  bfSize;
  unsigned short bfReserved1;
  unsigned short bfReserved2;
  unsigned long  bfOffBits;
} BITMAPFILEHEADER;

typedef struct tagBITMAPINFOHEADER{
    unsigned long  biSize;
    long           biWidth;
    long           biHeight;
    unsigned short biPlanes;
    unsigned short biBitCount;
    unsigned long  biCompression;
    unsigned long  biSizeImage;
    long           biXPixPerMeter;
    long           biYPixPerMeter;
    unsigned long  biClrUsed;
    unsigned long  biClrImporant;
} BITMAPINFOHEADER;

typedef struct tagGRPHEADER {
	unsigned short	mark;
	short	width;
	short	height;
	short	type;
	short	comp1;
	short	comp2;
	unsigned char yobi[20];
} GRPHEADER;

#pragma pack()

unsigned short *GRPbuff = NULL;
unsigned char ColorFlag[65536];

char	NoComp = 0;

//*******************************************************
//	BMP->GRP作成
//*******************************************************
int make_grp(char *bmp_fname, char *grp_fname){
	FILE	*bmp, *grp;
	BITMAPFILEHEADER	fh;
	BITMAPINFOHEADER	info;
	int		width, height;
	int		x, y, c;
	GRPHEADER	grp_header;
	unsigned short	*p, *ps;
	int		min;
	int		comp1;
	int		comp2 = 0;
	char	flag32 = 0;

	bmp = fopen(bmp_fname, "rb");
	if (bmp == NULL){
		fprintf(stderr, "file not open:%s\n", bmp_fname);
		return -1;
	}

	// BMPファイルヘッダ読み込み・チェック
	if (fread((void *)&fh, 1, sizeof(BITMAPFILEHEADER), bmp) != sizeof(BITMAPFILEHEADER)){
		fclose(bmp);
		fprintf(stderr, "file not read:%s\n", bmp_fname);
		return -1;
	}
	if (fh.bfType != 0x4D42){
		fclose(bmp);
		fprintf(stderr, "is not bmp file:%s\n", bmp_fname);
		return -1;
	}

	if (fread((void *)&info, 1, sizeof(BITMAPINFOHEADER), bmp) != sizeof(BITMAPINFOHEADER)){
		fclose(bmp);
		fprintf(stderr, "file not read:%s\n", bmp_fname);
		return -1;
	}
	if (info.biBitCount == 32){		// 32bitカラーか
		flag32 = 1;
	} else if (info.biBitCount != 24){		// 24bitカラーのみ
		fclose(bmp);
		fprintf(stderr, "color depth error:%s %d\n", bmp_fname, info.biBitCount);
		return -1;
	}
	if ((info.biWidth > WIDTH_MAX)||(info.biHeight > HEIGHT_MAX)){
		fclose(bmp);
		fprintf(stderr, "size over:%s %ldx%ld\n", bmp_fname, info.biWidth, info.biHeight);
		return -1;
	}


	// データ頭出し
	fseek(bmp, fh.bfOffBits, SEEK_SET);

	width = info.biWidth;
	height = info.biHeight;

	GRPbuff = (unsigned short *)malloc(width * height * sizeof(unsigned short));
	if (GRPbuff == NULL){
		fclose(bmp);
		fprintf(stderr, "malloc Error\n");
		return -1;
	}

	// BMPファイルは下から入っている
	for (y = height - 1; y >= 0; --y){
		unsigned char buff[4];
		for (x = 0; x < width; x++){
			unsigned short	r, g, b, color;

			if (flag32){
				if (fread(buff, 1, 4, bmp) != 4){
					fclose(bmp);
					fprintf(stderr, "file not read:%s\n", bmp_fname);
					return -1;
				}
				b = buff[1] >> 3;
				g = buff[2] >> 2;
				r = buff[3] >> 3;
			} else {
				if (fread(buff, 1, 3, bmp) != 3){
					fclose(bmp);
					fprintf(stderr, "file not read:%s\n", bmp_fname);
					return -1;
				}
				b = buff[0] >> 3;
				g = buff[1] >> 2;
				r = buff[2] >> 3;
			}
			color = (r << 11)|(g << 5)|b;
			GRPbuff[y * width + x] = color;
		}
		// 1ラインごとに 4byteアライメント
		if (!flag32 && ((width * 3) % 4)){
			fread(buff, 1, 4 - ((width * 3) % 4), bmp);
		}
	}
	fclose(bmp);

	if (NoComp){
		// 圧縮しない
		comp1 = 0;
	} else {
		// データで使われている色を調べる
		memset(ColorFlag, 0, sizeof(ColorFlag));
		p = GRPbuff;
		for (y = 0; y < height; y++){
			for (x = 0; x < width; x++){
				if (ColorFlag[*p] < 255){
					ColorFlag[*p]++;
				}
				p++;
			}
		}
		// 一番使用頻度の小さい色を圧縮マークとする
		min = 255;
		comp1 = 0;
		for (c = 1; c < 65535; c++){	// 0 と FFFF は除外
			if (ColorFlag[c] < min){
				min = ColorFlag[c];
				comp1 = c;
				if (min == 0)	break;
			}
		}
	}
	
	// GRPファイル作成
	memset(&grp_header, 0, sizeof(grp_header));
	grp_header.mark = 0x5047;		// マーク
	grp_header.width = width;		// 幅
	grp_header.height = height;		// 高さ
	grp_header.type = 1;			// タイプ(未使用)
	grp_header.comp1 = comp1;		// 圧縮マーク１
	grp_header.comp2 = comp2;		// 圧縮マーク２(未使用)
	
	grp = fopen(grp_fname, "wb");
	if (grp == NULL){
		fprintf(stderr, "file not write:%s\n", grp_fname);
		return -1;
	}
	if (fwrite(&grp_header, 1, sizeof(grp_header), grp) < sizeof(grp_header)){
		fclose(grp);
		fprintf(stderr, "write error:%s\n", grp_fname);
		return -1;
	}

	// ランレングス圧縮をしながらデータ書き込み
	ps = p = GRPbuff;
	while(p < &GRPbuff[width * height]){
		while(*(++ps) == *p){
			if (ps >= &GRPbuff[width * height])	break;
			if ((ps - p) >= 65534)	break;
		}
		if ((comp1 == 0)||			// 圧縮なし、もしくは、
			(((ps - p) <= 3)&&		// 繰り返し回数が３回以下 and
			 (*p != comp1))){		// 圧縮マークと色が一致していない
			int	i;

			for (i = 0; i < (ps - p); i++){
				if (fwrite(p, 2, 1, grp) < 1){
					fclose(grp);
					fprintf(stderr, "write error:%s\n", grp_fname);
					return -1;
				}
			}
		} else {
			unsigned short	buff[3];

			buff[0] = comp1;	// 圧縮マーク
			buff[1] = ps - p;	// 繰り返し回数
			buff[2] = *p;		// 色
			if (fwrite(buff, 2, 3, grp) < 3){
				fclose(grp);
				fprintf(stderr, "write error:%s\n", grp_fname);
				return -1;
			}
		}
		p = ps;
	}
	fclose(grp);
	
	return 0;	
}


//*******************************************************
//	メインルーチン
//*******************************************************
int main(int argc, char *argv[]){
	char	bmp_fname[1024];
	char	grp_fname[1024];
	char	*p;
	int		i;
	int		ret;

	memset(bmp_fname, 0, sizeof(bmp_fname));
	memset(grp_fname, 0, sizeof(grp_fname));
	
	for (i = 1; i < argc; i++){
		if (argv[i][0] == '-'){
			if ((strcmp(argv[i], "-nc") == 0)||
				(strncmp(argv[i], "--no-comp", 9) == 0)){
				NoComp = 1;	// 圧縮しない
			}
		} else {
			if (bmp_fname[0] == 0){
				strcpy(bmp_fname, argv[i]);
				p = strrchr(bmp_fname, '.');
				if (p == NULL){
					strcat(bmp_fname, ".bmp");
				}
			} else {
				strcpy(grp_fname, argv[i]);
				p = strrchr(grp_fname, '.');
				if (p == NULL){
					strcat(grp_fname, ".GRP");
				}
			}
		}
	}

	if (!grp_fname[0] && !bmp_fname[0]){
		printf("usage:\n");
		printf("\tmakegrp [option] BMPfile [GRPfile]\n");
		printf("\n");
		printf("option:\n");
		printf("\t-nc\t--no-comp\tNo Compress\n");
		return 0;
	}

	if (!grp_fname[0] && bmp_fname[0]){
		strcpy(grp_fname, bmp_fname);
		p = strrchr(grp_fname, '.');
		strcpy(p, ".GRP");
	}

	printf("%s -> %s\n", bmp_fname, grp_fname);

	ret = make_grp(bmp_fname, grp_fname);

	if (GRPbuff != NULL){
		free(GRPbuff);
	}

	return ret;
}


